#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Dict

COLORS: Dict[str, str] = {
    'primaryColor': '#94bbff',
    'primaryLightColor': '#94bbff',
    'secondaryColor': '#1e304f',
    'secondaryLightColor': '#061e3b',
    'secondaryDarkColor': '#4a597b',
    'primaryTextColor': '#ffffff',
    'secondaryTextColor': '#ffffff',
    'compoundColor': '#8dd35f',
    'reactionColor': '#ff9955',
    'associationColor': '#d4aa00',
    'flaskColor': '#5f8dd3',
    'structureColor': '#154734',
    'elementaryStepColor': '#a11f05',
    'borderColor': '#000000',
    'edgeColor': '#c8c8c8',
    'highlightColor': '#d40000',
}

MODE = 'dark'

LIGHT_MODES = ['light', 'legacy']
